/*
 * tine_history.cpp : 
 *
 */

#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define mxerr(s) { mxerrStr = (char *)s; goto err; }
#define ARCHIVE_CAPACITY 8000
DBLDBLDBL dddDataArray[ARCHIVE_CAPACITY];
DBLDBL *ddDataArray = (DBLDBL *)dddDataArray;
double *dDataArray = (double *)dddDataArray;
BYTE* bDataArray = (BYTE *)dddDataArray;

int read_prop(char *dev, char *prop, int *size, char *tag)
{ 
  PropertyQueryStruct	svrProp; // legacy
  PrpQueryStruct pqs;
  DTYPE dout, din;
  int cc, fmt, isLegacy=FALSE;
  char *prp = "PROPS";

  if ((cc=getPropertyInfoFromListenerEx(dev,prop,size,&fmt,tag)) == 0)
  {
    return fmt;
  }
  DTYPEZERO(din); DTYPEZERO(dout);
  din.dFormat = CF_TEXT;
  din.dArrayLength = (int)strlen(prop);
  din.data.vptr = prop;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  dout.data.vptr = &pqs;
  strcpy(dout.dTag,"PRPQSr4");
legacyRetry:
  if ((cc=alsnExecLink(dev,prp,&dout,&din,CA_READ,500)) != 0)
  {
    if (!isLegacy && (cc == illegal_format || cc == invalid_structure_tag))
    {
      isLegacy = TRUE;
      dout.dFormat = CF_STRUCT;
      dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
      dout.data.vptr = &svrProp;
      dout.dTag[0] = '\0';
      goto legacyRetry;
    }
    *size = 0;
    return -cc;
 	}
  if (isLegacy)
  {
    *size = svrProp.prpSize;
    fmt = LFMT(svrProp.prpFormat);
  }
  else
  {
    *size = pqs.prpSize;
    fmt = LFMT(pqs.prpFormat);
    if (tag != NULL) strncpy(tag,pqs.prpTag,TAG_NAME_SIZE);
  }
  return fmt;
}
time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[16];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  hr[0] = 0;
  tnow = localtime(&t);
  len = (int)strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year,"_")) != NULL || (c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
	if (tmts.tm_year < 100) tmts.tm_year += 2000;
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
int mlParseFullDeviceName(char *fullname,char *dev,char *prp,int isHistoryCall)
{
  char *c,tmp[STATUS_SIZE+8];
  int isBracketed = FALSE;

  strncpy(tmp,fullname,STATUS_SIZE+8);
  if ((c=strchr(tmp,'[')) != 0)
  { /* fullname given as "/<context>/<server>/<device>[<property>]" */
    *c++ = 0;
    strncpy(prp,c,PROPERTY_NAME_SIZE);
    if ((c=strchr(prp,']')) != 0) *c = 0;
    isBracketed = TRUE;
  }
  if (!isBracketed)
  { /* fullname given as "/<context>/<server>/<device>/<property>" */
    strrev(tmp);
    if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
    *c++ = 0;
    strncpy(prp,tmp,PROPERTY_NAME_SIZE); strrev(prp);
    strncpy(dev,c,STATUS_SIZE); strrev(dev);
  }
  else
  {
    strncpy(dev,tmp,STATUS_SIZE);
  }
  if (isHistoryCall)
  {
    if (strstr(dev,"HISTORY") == NULL && /* not a central archive call */
      strstr(prp,".HIST") == NULL && strstr(prp,".HST") == NULL) /* meta property not yet appended */
    {
      if (strlen(prp) < PROPERTY_NAME_SIZE-4) strcat(prp,".HST");
    }
  }
  return 0;
}
int mlGetArchivedDataAsSnapshot(char *devsrv,time_t *target,double *dataSnapshot,int dataType,int *len)
{
  DTYPE din,dout;
  UINT32 snapshotArray[4];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (len == NULL || *len <= 0) return buffer_too_small;
  if (target == NULL) return argument_list_error;
  if ((cc=mlParseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  DTYPEZERO(dout); DTYPEZERO(din);
  snapshotArray[0] = (UINT32)*target;
  snapshotArray[1] = (UINT32)*target;
  dout.dArrayLength = *len;
  dout.dFormat = GetNumberSegments(dataType) == 1 ? CF_DOUBLE : dataType;
  if (GetFormatHeaderSize(dataType) > 0) dout.dFormat = dataType;
  dout.data.vptr = dataSnapshot;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = snapshotArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,200)) != 0) 
  {
    *len = 0;
  }
  else 
  {
    *len = (int)dout.dArrayLength;
	  *target = (UINT32)dout.dTimeStamp;
  }
  return cc;
}
int mlGetArchivedDataAsDoublet(char *devsrv,time_t start,time_t stop,int sample,int index,DBLDBL *ddDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[4];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=mlParseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = (UINT32)index;
  startstopArray[3] = (UINT32)sample;
  dout.dArrayLength = *num;
  dout.dFormat = CF_DBLDBL;
  dout.data.vptr = ddDataArray;
  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,200)) != 0) *num = 0;
  else *num = (int)dout.dArrayLength;

  return cc;
}
int mlGetArchivedDataAsTriplet(char *devsrv,time_t start,time_t stop,int sample,int index,DBLDBLDBL *dddDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[4];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=mlParseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = (UINT32)index;
  startstopArray[3] = (UINT32)sample;
  dout.dArrayLength = *num;
  dout.dFormat = CF_DBLDBLDBL;
  dout.data.vptr = ddDataArray;
  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,200)) != 0) *num = 0;
  else *num = (int)dout.dArrayLength;

  return cc;
}
#define usageStr "tine_history usage:\ntine_history('/<context>/<server>/<name>[<property>]','<stop time>','<depth>'[,sampleInterval,index,'acquireSystemStamps'])\n'ret' contains array of (timestamp,data) or (timestamp,system stamp,data)"
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  char *parameter, *mxerrStr=NULL;
  time_t start, stop, dt;
  int type, num, idx;
  mxArray *temp;
  double *ptr;
  int buflen, status, cc=argument_list_error;
  char *c, c0, str[128], dtstr[32];
  int i,len,depth=86400,sample=0,acqSysStmp=FALSE,acqSnapshot=FALSE;
 
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 6:
      if (mxIsChar(prhs[5])) 
      {
        len = (int)mxGetN(prhs[5]) + 1;
        c = (char *)mxCalloc(len, sizeof(char));
        mxGetString(prhs[5],c,len);
        c0 = toupper(*c);
        if (c0 == 'Y' || c0 == 'T') acqSysStmp = TRUE;
      }
      else if (mxIsDouble(prhs[5])) 
      {
        acqSysStmp = (int)mxGetScalar(prhs[5]);
      }
    case 5:
      if (!mxIsDouble(prhs[4])) mxerr(usageStr);
      idx = (int)mxGetScalar(prhs[4]);
    case 4:
      if (!mxIsDouble(prhs[3])) mxerr(usageStr);
      sample = (int)mxGetScalar(prhs[3]);
    case 3:
      if (mxIsChar(prhs[2]) != 1) mxerr("depth input must be a string.");
      mxGetString(prhs[2],str,128);
      if (!strncmp(str,"SNAP",4) || !strncmp(str,"snap",4) || !strncmp(str,"Snap",4))
      {
        acqSnapshot = TRUE;
      }
      else
      {
        depth = strtol(str,&c,10);
        switch (tolower(*c))
        {
          case 'd': depth *= 86400; 
            break;
          case 'h': depth *= 3600; 
            break;
          case 'm':
            if (strnicmp(c,"min",3) == 0)
            {
              depth *= 60;
            }
            else if (strnicmp(c,"mon",3) == 0)
            {
              depth *= 86400 * 31;
            }
            break;
          case 'w':
            depth *= 86400 * 7;
            break;
          default:
          case 's': 
            break;
        }
      }
    case 2:
      if (mxIsChar(prhs[1]) != 1) mxerr("stop time input must be a string.");
      mxGetString(prhs[1],str,128);
      stop = parseTimeString(str);
      if (acqSnapshot)
      {
        start = stop;
        break;
      }
      start = stop - depth;
      if (mxGetM(prhs[0]) != 1) mxerr("Input must be a row vector.");
      if (mxIsChar(prhs[0]) != 1) mxerr("target input must be a string.");
      break;
    default:
      mxerr(usageStr);
  }
  if (nlhs > 1) mxerr("Too many output arguments");

  /* Get the length of the input string. */
  buflen = mxGetM(prhs[0])*mxGetN(prhs[0]) + 1;
  
  /* Allocate memory for input and output strings. */
  parameter = (char *) mxCalloc(buflen, sizeof(char));
  
  /* copy string into parameter */
  status = mxGetString(prhs[0], parameter, buflen);
  if(status != 0) mexWarnMsgTxt("Not enough space. String is truncated.");
	SetUserType("MatLab");

  num = ARCHIVE_CAPACITY;
  if (acqSnapshot)
  {
    char dtag[32];
    char dev[192], *prop, *c;
    strncpy(dev,parameter,192);
    if ((prop=strchr(dev,'[')) != NULL) 
    {
      *prop++ = 0;
      if ((c=strchr(prop,']')) != NULL) *c = 0;
      type = read_prop(dev, prop, &num, dtag);
      if (num > ARCHIVE_CAPACITY*3) num = ARCHIVE_CAPACITY*3;
	    dt = start;
      cc = mlGetArchivedDataAsSnapshot(parameter,&dt,dDataArray,type,&num);
    }
  }
  else if (acqSysStmp)
  {
    cc = mlGetArchivedDataAsTriplet(parameter, start, stop, sample, idx, dddDataArray, &num);
  }
  else
  {
    cc = mlGetArchivedDataAsDoublet(parameter, start, stop, sample, idx, ddDataArray, &num);
  }
  if (cc != 0) 
  {
  	mexPrintf("Error reading data, status = %d.\n", cc);
    mexPrintf("Error code is: %s\n", GetRPCLastError(cc));
    num = 0;
  }
  /* create matrix for data */
  if (acqSnapshot)
  {
    mxArray *dtemp;
    int nsegs = GetNumberSegments(type);
    int hsiz = GetFormatHeaderSize(type);
    dtemp = mxCreateDoubleMatrix(num, nsegs, mxREAL);
    const char *field[3];
	  char tsstr[64];
    field[0] = (char *)mxCalloc(10, sizeof(char));
    field[1] = (char *)mxCalloc(4, sizeof(char));
    field[2] = (char *)mxCalloc(6, sizeof(char));
    memset((void *)field[0],0,sizeof(field[0]));
    memset((void *)field[1],0,sizeof(field[1]));
    memset((void *)field[2],0,sizeof(field[2]));
    strcpy((char *)field[0],"timestamp");
    strcpy((char *)field[1],"utc");
    strcpy((char *)field[2],"data");
    temp = mxCreateStructMatrix(1,1,3,&field[0]);
    mxArray *data[3];
    strncpy(tsstr,getDataTimeString(dt,0),64);
    sprintf(dtstr,"%d",dt);
    data[0] = mxCreateString(tsstr);
    data[1] = mxCreateString(dtstr);
	  data[2] = dtemp;
    mxSetFieldByNumber(temp,0,0,data[0]);
    mxSetFieldByNumber(temp,0,1,data[1]);
    mxSetFieldByNumber(temp,0,2,data[2]);
    ptr = mxGetPr(dtemp);
    switch (nsegs)
    {
      default:
      case 1:
        if (hsiz > 0)
        {
          float *fvals = (float *)&bDataArray[SPECTRUM_HDR_SIZE];
          if (type != CF_SPECTRUM)
          {
            mexPrintf("Data tyoe not supported for archive calls\n", cc);
            num = 0;
            goto err;
          }
          dt = (double)((hdrSPECTRUM *)bDataArray)->tm;
          strncpy(tsstr, getDataTimeString(dt, 0), 64);
          sprintf(dtstr, "%d", dt);
          for (i = 0; i < num; i++)
          {
            ptr[i] = fvals[i];
          }
        }
        else
        {
          for (i = 0; i < num; i++) ptr[i] = dDataArray[i];
        }
        break;
      case 2:
        switch (type)
        {
          case CF_FLTFLT:
          {
            FLTFLT *ff = (FLTFLT *)dDataArray;
            for (i=0; i<num; i++) 
            {
              ptr[i] = ff[i].f1val;
              ptr[num+i] = ff[i].f2val;
            }
            break;
          }
          case CF_DBLDBL:
          {
            for (i=0; i<num; i++) 
            {
              ptr[i] = ddDataArray[i].d1val;
              ptr[num+i] = ddDataArray[i].d2val;
            }
            break;
          }
          case CF_FLTINT:
          {
            FLTINT *fi = (FLTINT*)dDataArray;
            for (i=0; i<num; i++) 
            {
              ptr[i] = fi[i].fval;
              ptr[num+i] = fi[i].ival;
            }
            break;
          }
          case CF_INTINT:
          {
            INTINT *ii = (INTINT*)dDataArray;
            for (i=0; i<num; i++) 
            {
              ptr[i] = ii[i].i1val;
              ptr[num+i] = ii[i].i2val;
            }
            break;
          }
          default:
            break;
        }
        break;      
    }
  }
  else if (acqSysStmp)
  {
    temp = mxCreateDoubleMatrix(num, 3, mxREAL);
    ptr = mxGetPr(temp);
    for (i = 0; i < num; i++) {
        ptr[i]     = dddDataArray[i].d2val/86400.0 + 719529.0;
        ptr[i+num] = dddDataArray[i].d3val;
        ptr[i+2*num] = dddDataArray[i].d1val;
    }
  }
  else
  {
    temp = mxCreateDoubleMatrix(num, 2, mxREAL);
    ptr = mxGetPr(temp);
    for (i = 0; i < num; i++) {
        ptr[i]     = ddDataArray[i].d2val/86400.0 + 719529.0;
        ptr[i+num] = ddDataArray[i].d1val;
    }
  }
  plhs[0] = temp; 
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (parameter != NULL) mxFree(parameter);
}